/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecenc.c                                               *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-09 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecenc.c 51 2006-07-09 16:30:19Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "leco.h"
#include "crc.h"
#include "block.h"
#include "gf.h"
#include "lecenc.h"
#include "lecofile.h"
#include "md5.h"
#include "util.h"


typedef struct {
	lec_file_t    lf;

	FILE          *inp;

	unsigned      bufcnt;
	unsigned char *buf;
	uint16_t      **blk;

	md5_t         md5;
	crc32_t       fcrc;
	crc32_t       dcrc;
} lec_enc_t;


int lec_enc_init (lec_enc_t *le, FILE *inp, FILE *out)
{
	lecf_init (&le->lf, out);

	le->inp = inp;

	le->bufcnt = 0;

	le->buf = NULL;
	le->blk = NULL;

	md5_init (&le->md5);
	crc32_init (&le->fcrc, CRC_CRC32);
	crc32_init (&le->dcrc, CRC_CRC32);

	return (0);
}

void lec_enc_free (lec_enc_t *le)
{
	free (le->buf);

	if (le->blk != NULL) {
		free (le->blk[0]);
		free (le->blk);
	}
}

int lec_enc_alloc (lec_enc_t *le, unsigned cnt)
{
	unsigned i;

	le->bufcnt = cnt;

	le->buf = malloc (le->lf.size + 8);

	le->blk = malloc (cnt * sizeof (uint16_t *));
	le->blk[0] = malloc (cnt * le->lf.bsize * sizeof (uint16_t));

	for (i = 1; i < cnt; i++) {
		le->blk[i] = le->blk[i - 1] + le->lf.bsize;
	}

	return (0);
}

int lec_enc_md5 (lec_enc_t *le)
{
	unsigned i;

	md5_done (&le->md5);
	md5_get_sum (&le->md5, le->lf.md5);

	crc32_done (&le->fcrc, 1);
	le->lf.fcrc = crc32_get_reg (&le->fcrc);

	crc32_done (&le->dcrc, 1);
	le->lf.dcrc = crc32_get_reg (&le->dcrc);

	printf ("crc:             %08lx (%08lx)\n", le->lf.fcrc, le->lf.dcrc);

	if (lecf_write_header (&le->lf)) {
		return (1);
	}

	printf ("md5:             ");

	for (i = 0; i < 16; i++) {
		printf ("%02x", le->lf.md5[i]);
	}

	printf ("\n");

	return (0);
}

int lec_enc_block (lec_enc_t *le, unsigned srci, unsigned srcn, unsigned dsti, unsigned dstn)
{
	unsigned      i, j;
	unsigned long cnt;
	uint64_t      blki, blkn;
	uint64_t      fidx, oidx, iidx;
	uint16_t      v;
	unsigned      pct1, pct2;
	lec_file_t    *lf;

	lf = &le->lf;

	/* blki: number of blocks already processed, blkn: total number of blocks */
	blki = srci * lf->ecnt + (srcn - srci) * dsti;
	blkn = lf->bcnt * lf->ecnt;
	pct1 = 1001;

	dstn = dstn - dsti;

	fidx = (uint64_t) srci * lf->size;
	oidx = lf->eofs + (uint64_t) dsti * (lf->size + 8);
	iidx = oidx;

	if (srci == 0) {
		for (j = 0; j < dstn; j++) {
			lec_block_clear (le->blk[j + 1], lf->bsize);
		}
	}
	else {
		for (j = 0; j < dstn; j++) {
			if (lec_read (lf->fp, iidx + 4, le->buf, lf->size) != lf->size) {
				return (1);
			}

			lec_block_get (le->blk[j + 1], le->buf, lf->bsize);

			iidx += lf->size + 8;
		}
	}

	for (i = srci; i < srcn; i++) {
		if (blkn == 0) {
			pct2 = 1000;
		}
		else {
			pct2 = (1000ULL * blki) / blkn;
		}

		if (pct2 != pct1) {
			printf ("encoding blocks: %u.%u%%  \r", pct2 / 10, pct2 % 10);
			fflush (stdout);
			pct1 = pct2;
		}

		cnt = lec_read (le->inp, fidx, le->buf, lf->size);

		if (cnt != lf->size) {
			if (i < (lf->bcnt - 1)) {
				printf ("\nread error\n");
				return (1);
			}

			for (j = cnt; j < lf->size; j++) {
				le->buf[j] = 0;
			}
		}

		lec_block_get (le->blk[0], le->buf, lf->bsize);

		if (dsti == 0) {
			lec_set_uint32 (le->buf, lf->size, lec_get_crc (le->buf, lf->size));

			md5_calc (&le->md5, le->buf, cnt);
			crc32_calc (&le->fcrc, le->buf, cnt);
			crc32_calc (&le->dcrc, le->buf + lf->size, 4);

			if (lec_write (lf->fp, lf->dofs + 4UL * i, le->buf + lf->size, 4) != 4) {
				return (1);
			}
		}

		for (j = 0; j < dstn; j++) {
			v = lec_get_coeff (par_first_block + dsti + j, i);
			lec_block_muladd (le->blk[j + 1], le->blk[0], v, lf->bsize);
		}

		blki += dstn;
		fidx += cnt;
	}

	if (dsti == 0) {
		/* about to write the first ecc block */
		if (lecf_write_magic2 (lf)) {
			return (1);
		}
	}

	for (j = 0; j < dstn; j++) {
		uint32_t crc;

		lec_set_uint32 (le->buf, 0, par_first_block + dsti + j);
		lec_block_set (le->buf + 4, le->blk[j + 1], lf->bsize);
		crc = lec_get_crc (le->buf, lf->size + 4);
		lec_set_uint32 (le->buf, lf->size + 4, crc);

		if (lec_write (lf->fp, oidx, le->buf, lf->size + 8) != (lf->size + 8)) {
			return (1);
		}

		oidx += lf->size + 8;
	}

	return (0);
}

int lec_encode_le (lec_enc_t *le)
{
	unsigned      si, sn, ss;
	unsigned      di, dn, ds;
	unsigned long tmp;
	lec_file_t    *lf;

	lf = &le->lf;

	tmp = par_src_size / lf->size;
	if (tmp == 0) {
		tmp = 1;
	}
	else if (tmp > lf->bcnt) {
		tmp = lf->bcnt;
	}

	ss = tmp;

	tmp = par_dst_size / lf->size;
	if (tmp == 0) {
		tmp = 1;
	}
	else if (tmp > lf->ecnt) {
		tmp = lf->ecnt;
	}
	ds = tmp;

	printf ("encode chunks:   %u * %u blocks\n", ss, ds);

	if (lec_enc_alloc (le, 1 + ds)) {
		return (1);
	}

	if (lecf_write_header (lf)) {
		return (1);
	}

	if (lecf_write_magic1 (lf)) {
		return (1);
	}

	si = 0;
	while (si < lf->bcnt) {
		sn = (lf->bcnt - si);
		if (sn > ss) {
			sn = ss;
		}

		if (lf->ecnt == 0) {
			if (lec_enc_block (le, si, si + sn, 0, 0)) {
				return (1);
			}
		}

		di = 0;
		while (di < lf->ecnt) {
			dn = (lf->ecnt - di);
			if (dn > ds) {
				dn = ds;
			}

			if (lec_enc_block (le, si, si + sn, di, di + dn)) {
				return (1);
			}

			di += dn;
		}

		si += sn;
	}

	printf ("encoding blocks: 100.0%%  \n");
	fflush (stdout);

	if (lec_enc_md5 (le)) {
		return (1);
	}

	if (par_redundant_crcs) {
		if (lecf_write_redundant (lf, NULL)) {
			return (1);
		}

		printf ("redundant crcs:  done\n");
	}

	return (0);
}

int lec_enc_get_params (lec_enc_t *le, uint64_t fsize)
{
	unsigned long esize;
	unsigned long eccover;
	lec_file_t    *lf;

	lf = &le->lf;

	if ((par_size == 0) && (par_bcnt == 0)) {
		printf ("need a data block size\n");
		return (1);
	}

	lf->fsize = fsize;

	if (par_bcnt != 0) {
		lf->bsize = ((lf->fsize + 1) / 2 + par_bcnt - 1) / par_bcnt;
		lf->size = 2 * lf->bsize;
	}
	else {
		lf->bsize = par_size / 2;
		lf->size = 2 * lf->bsize;
	}

	if (lf->bsize < 32) {
		lf->bsize = 32;
		lf->size = 64;
	}

	lf->bcnt = (lf->fsize + lf->size - 1) / lf->size;

	if (lf->bcnt > 16384) {
		printf ("too many data blocks\n");
		return (1);
	}

	eccover = (LECO_HEADER + 15) & ~15UL;
	eccover += 16;
	eccover += 4UL * lf->bcnt;
	eccover = (eccover + 15) & ~15UL;
	eccover += 16;

	if (par_redundant_crcs) {
		eccover += 16 + 32;
		eccover += 4UL * lf->bcnt;
	}

	if (par_ecnt != 0) {
		lf->ecnt = par_ecnt;
	}
	else {
		if (par_esize2 != 0) {
			if (par_esize2 >= lf->fsize) {
				esize = par_esize2 - lf->fsize;
			}
			else {
				esize = 0;
			}
		}
		else {
			esize = par_esize;
		}

		if (esize < eccover) {
			lf->ecnt = 0;
		}
		else {
			lf->ecnt = ((esize - eccover) & ~15UL) / (lf->size + 8);
		}
	}

	if (par_redundant_crcs) {
		esize = ((lf->size + 8) * lf->ecnt + 15) & ~15UL;
	}
	else {
		esize = (lf->size + 8) * lf->ecnt;
	}

	lecf_calc_dofs (lf);
	lecf_calc_eofs (lf);
	lecf_calc_rofs (lf);

	printf ("block size:      %lu\n", lf->size);
	printf ("data size:       %-9llu (%u blocks)\n",
		(unsigned long long) lf->fsize, lf->bcnt
	);
	printf ("ecc size:        %-9lu (%u blocks)\n", eccover + esize, lf->ecnt);
	printf ("first ecc block: %u\n", par_first_block);

	if (lf->ecnt > 65535) {
		printf ("too many ecc blocks\n");
		return (1);
	}

	return (0);
}

int lec_encode_fp (FILE *inp, FILE *out, unsigned long long fsize)
{
	int       r;
	lec_enc_t le;

	if (fsize == 0) {
		if (fseeko (inp, 0, SEEK_END)) {
			return (1);
		}

		fsize = ftello (inp);
	}

	if (lec_enc_init (&le, inp, out)) {
		return (1);
	}

	if (lec_enc_get_params (&le, fsize)) {
		return (1);
	}

	r = lec_encode_le (&le);

	lec_enc_free (&le);

	return (r);
}

int lec_encode (const char *finp, const char *fout)
{
	int                r;
	FILE               *inp, *out;
	unsigned long long fsize;
	struct stat        st;

	if (stat (finp, &st)) {
		fprintf (stderr, "%s: can't stat input file (%s)\n", argv0, finp);
		fsize = 0;
	}
	else {
		fsize = st.st_size;
	}

	inp = fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
		return (1);
	}

	out = fopen (fout, "w+b");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
		fclose (inp);
		return (1);
	}

	r = lec_encode_fp (inp, out, fsize);

	fclose (out);
	fclose (inp);

	return (r);
}
