/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/leco.c                                                 *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-08 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: leco.c 48 2006-07-08 11:37:23Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "leco.h"
#include "block.h"
#include "crc.h"
#include "gf.h"
#include "lecdec.h"
#include "lecenc.h"
#include "strtoint.h"
#include "util.h"


const char    *argv0 = NULL;

int           par_encode = 1;

uint64_t      par_esize = 0;
uint64_t      par_esize2 = 0;
unsigned      par_ecnt = 1;

unsigned long par_size = 0;
unsigned      par_bcnt = 256;

unsigned      par_first_block = 0;

unsigned long par_src_size = 64UL * 1024UL * 1024UL;
unsigned long par_dst_size = 4UL * 1024UL * 1024UL;

int           par_redundant_crcs = 0;


static
void prt_version (void)
{
	fputs (
		"leco version " LECO_VERSION_STR
		" (" LECO_CFG_DATE " " LECO_CFG_TIME ")\n\n"
		"Copyright (C) 2005 by Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_help (void)
{
	fputs (
		"Usage: leco [options] [input ecc-file] [output]\n"
		"  -d, --decode            Repair the input file\n"
		"  -e, --encode            Create the ecc file\n"
		"  -i string               Set input file name\n"
		"  -o string               Set output file name\n"
		"  -l string               Set the ecc file name\n"
		"  -b, --ecc-chunk int     Set ecc chunk size [1M]\n"
		"  -B, --data-chunk int    Set data chunk size [256M]\n"
		"  -f, --first-block int   Set the forst ecc block index [0]\n"
		"  -n, --data-blocks int   Set the number of data blocks [256]\n"
		"  -N, --block-size int    Set the block size\n"
		"  -s, --ecc-size int      Set the total ecc size\n"
		"  -S, --ecc-blocks int    Set the number of ecc blocks [1]\n"
		"  -t, --total-size int    Set data size + ecc size\n"
		"  -r, --redundant-crcs    Add redundant data crcs [no]\n",
		stdout
	);

	fflush (stdout);
}

int main (int argc, char **argv)
{
	int        i;
	const char *f1, *f2, *f3;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg (argv[1], NULL, "--version")) {
			prt_version();
			return (0);
		}
		else if (str_isarg (argv[1], NULL, "--help")) {
			prt_help();
			return (0);
		}
	}

	f1 = NULL;
	f2 = NULL;
	f3 = NULL;

	i = 1;
	while (i < argc) {
		if (str_isarg (argv[i], "-d", "--decode")) {
			par_encode = 0;
		}
		else if (str_isarg (argv[i], "-e", "--encode")) {
			par_encode = 1;
		}
		else if (str_isarg (argv[i], "-i", "--input")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (1);
			}

			f1 = argv[i];
		}
		else if (str_isarg (argv[i], "-l", "--ecc")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing ecc file name\n", argv0);
				return (1);
			}

			f2 = argv[i];
		}
		else if (str_isarg (argv[i], "-o", "--output")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (1);
			}

			f3 = argv[i];
		}
		else if (str_isarg (argv[i], "-b", "--ecc-chunk")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing ecc chunk size\n", argv0);
				return (1);
			}

			if (str_get_ulng (argv[i], &par_dst_size)) {
				fprintf (stderr, "%s: bad ecc chunk size (%s)\n", argv0, argv[i]);
				return (1);
			}
		}
		else if (str_isarg (argv[i], "-f", "--first-block")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing first block index\n", argv0);
				return (1);
			}

			if (str_get_uint (argv[i], &par_first_block)) {
				fprintf (stderr, "%s: bad first block index (%s)\n", argv0, argv[i]);
				return (1);
			}
		}
		else if (str_isarg (argv[i], "-B", "--data-chunk")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing data chunk size\n", argv0);
				return (1);
			}

			if (str_get_ulng (argv[i], &par_src_size)) {
				fprintf (stderr, "%s: bad data chunk size (%s)\n", argv0, argv[i]);
				return (1);
			}
		}
		else if (str_isarg (argv[i], "-n", "--data-blocks")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing number of data blocks\n", argv0);
				return (1);
			}

			if (str_get_uint (argv[i], &par_bcnt)) {
				fprintf (stderr, "%s: bad number of data blocks (%s)\n", argv0, argv[i]);
				return (1);
			}

			par_size = 0;
		}
		else if (str_isarg (argv[i], "-N", "--block-size")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing data block size\n", argv0);
				return (1);
			}

			if (str_get_ulng (argv[i], &par_size)) {
				fprintf (stderr, "%s: bad data block size (%s)\n", argv0, argv[i]);
				return (1);
			}

			par_bcnt = 0;
		}
		else if (str_isarg (argv[i], "-r", "--redundant-crcs")) {
			par_redundant_crcs = 1;
		}
		else if (str_isarg (argv[i], "-s", "--ecc-size")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing ecc size\n", argv0);
				return (1);
			}

			if (str_get_uint64 (argv[i], &par_esize)) {
				fprintf (stderr, "%s: bad ecc size (%s)\n", argv0, argv[i]);
				return (1);
			}

			par_esize2 = 0;
			par_ecnt = 0;
		}
		else if (str_isarg (argv[i], "-t", "--total-size")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing total size\n", argv0);
				return (1);
			}

			if (str_get_uint64 (argv[i], &par_esize2)) {
				fprintf (stderr, "%s: bad total size (%s)\n", argv0, argv[i]);
				return (1);
			}

			par_esize = 0;
			par_ecnt = 0;
		}
		else if (str_isarg (argv[i], "-S", "--ecc-blocks")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing number of ecc blocks\n", argv0);
				return (1);
			}

			if (str_get_uint (argv[i], &par_ecnt)) {
				fprintf (stderr, "%s: bad number of ecc blocks (%s)\n", argv0, argv[i]);
				return (1);
			}

			par_esize = 0;
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n", argv[0], argv[i]);
			return (1);
		}
		else {
			if (f1 == NULL) {
				f1 = argv[i];
			}
			else if (f2 == NULL) {
				f2 = argv[i];
			}
			else if (f3 == NULL) {
				f3 = argv[i];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv[0], argv[i]);
				return (1);
			}
		}

		i += 1;
	}

	if (gf_init (&par_gf, 16)) {
		return (1);
	}

	if (par_encode) {
		if (lec_encode (f1, f2)) {
			return (1);
		}
	}
	else {
		if (lec_decode (f1, f2, f3)) {
			return (1);
		}
	}

	gf_free (&par_gf);

	return (0);
}
