/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecocat.c                                              *
 * Created:       2006-07-04 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-08 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2006 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecocat.c 48 2006-07-08 11:37:23Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"
#include <unistd.h>

#include "lecofile.h"
#include "util.h"


typedef struct {
	unsigned long idx;
	FILE          *fp;
	uint64_t      ofs;
} lec_cat_blk_t;


typedef struct {
	char       *fname;
	lec_file_t lf;
} lec_cat_file_t;


const char            *argv0 = NULL;

static int            par_verbose = 0;

static unsigned long  par_blk_cnt = 0;
static unsigned long  par_blk_max = 0;
static unsigned long  par_blk_total = 0;
static lec_cat_blk_t  *par_blk = NULL;

static unsigned       par_lec_cnt = 0;
static lec_cat_file_t *par_lec = NULL;


static
void prt_version (void)
{
	fputs (
		"lecocat version " LECO_VERSION_STR
		" (" LECO_CFG_DATE " " LECO_CFG_TIME ")\n\n"
		"Copyright (C) 2006 by Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_help (void)
{
	fputs (
		"Usage: lecocat [options] [files...]\n"
		"  -v, --verbose  Verbose operation [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int lec_cat_compare (lec_file_t *lf1, lec_file_t *lf2)
{
	if (lf1->bsize != lf2->bsize) {
		return (1);
	}

	if (lf1->size != lf2->size) {
		return (1);
	}

	if (lf1->wsize != lf2->wsize) {
		return (1);
	}

	if (lf1->bcnt != lf2->bcnt) {
		return (1);
	}

	if (lf1->fsize != lf2->fsize) {
		return (1);
	}

	if (memcmp (lf1->md5, lf2->md5, 16) != 0) {
		return (1);
	}

	return (0);
}

/* make sure all input files correspond to the same data file */
static
int lec_cat_check_files (void)
{
	unsigned       i;
	lec_cat_file_t *lc1, *lc2;

	lc1 = par_lec;

	for (i = 1; i < par_lec_cnt; i++) {
		lc2 = par_lec + i;

		if (lec_cat_compare (&lc1->lf, &lc2->lf)) {
			fprintf (stderr, "%s: %s: bad header\n", argv0, lc2->fname);
			return (1);
		}
	}

	return (0);
}

/* check if block idx is already in the list */
static
int lec_cat_check_block (unsigned long idx)
{
	unsigned long i;

	for (i = 0; i < par_blk_cnt; i++) {
		if (par_blk[i].idx == idx) {
			return (1);
		}
	}

	return (0);
}

static
int lec_cat_add_block (unsigned long idx, FILE *fp, uint64_t ofs)
{
	unsigned long i;

	if (par_blk_cnt >= par_blk_max) {
		unsigned long max;
		lec_cat_blk_t *tmp;

		max = (par_blk_max == 0) ? 1 : par_blk_max;

		while (max <= par_blk_cnt) {
			max *= 2;
		}

		tmp = realloc (par_blk, max * sizeof (lec_cat_blk_t));
		if (tmp == NULL) {
			return (1);
		}

		par_blk = tmp;
		par_blk_max = max;
	}

	i = par_blk_cnt;

	while ((i > 0) && (idx < par_blk[i - 1].idx)) {
		par_blk[i] = par_blk[i - 1];

		i -= 1;
	}

	par_blk[i].idx = idx;
	par_blk[i].fp = fp;
	par_blk[i].ofs = ofs;

	par_blk_cnt += 1;

	return (0);
}

static
int lec_cat_add_file_blocks (lec_cat_file_t *lc)
{
	unsigned long i;
	unsigned long idx;
	unsigned long cnt;
	uint32_t      crc1, crc2;
	uint64_t      ofs;
	unsigned char *buf;
	lec_file_t    *lf;

	lf = &lc->lf;

	if (par_verbose) {
		fprintf (stderr, "Processing file \"%s\"\n", lc->fname);
	}

	ofs = lf->eofs;
	cnt = 0;

	for (i = 0; i < lf->ecnt; i++) {
		buf = lec_read_buf (lf->fp, ofs, lf->size + 8);
		if (buf == NULL) {
			return (1);
		}

		idx = lec_get_uint32 (buf, 0);
		crc1 = lec_get_uint32 (buf, lf->size + 4);
		crc2 = lec_get_crc (buf, lf->size + 4);

		free (buf);

		if (crc1 == crc2) {
			if (lec_cat_check_block (idx)) {
				if (par_verbose) {
					fprintf (stderr, "    duplicate block: %lu\n", idx);
				}
			}
			else {
				if (lec_cat_add_block (idx, lf->fp, ofs)) {
					return (1);
				}

				cnt += 1;
			}
		}
		else {
			if (par_verbose) {
				fprintf (stderr, "    bad block: %lu\n", idx);
			}
		}

		ofs += lf->size + 8;
	}

	par_blk_total += lf->ecnt;

	if (par_verbose) {
		fprintf (stderr, "  blocks: %lu of %u\n", cnt, lf->ecnt);
	}

	return (0);
}

static
int lec_cat_add_blocks (void)
{
	unsigned i;

	for (i = 0; i < par_lec_cnt; i++) {
		if (lec_cat_add_file_blocks (&par_lec[i])) {
			fprintf (stderr, "failed\n");
			return (1);
		}
	}

	if (par_verbose) {
		fprintf (stderr, "blocks: %lu of %lu\n", par_blk_cnt, par_blk_total);
	}

	return (0);
}

static
int lec_cat_output_blocks (lec_file_t *lf)
{
	unsigned long i;
	unsigned long cnt;
	lec_cat_blk_t *blk;
	unsigned char *buf;
	uint64_t      ofs;

	lecf_write_magic2 (lf);

	blk = par_blk;
	ofs = lf->eofs;
	cnt = lf->size + 8;

	for (i = 0; i < lf->ecnt; i++) {
		buf = lec_read_buf (blk->fp, blk->ofs, cnt);
		if (buf == NULL) {
			fprintf (stderr, "%s: read error (block)\n", argv0);
			return (1);
		}

		if (lec_write (lf->fp, ofs, buf, cnt) != cnt) {
			fprintf (stderr, "%s: write error (block)\n", argv0);
			free (buf);
			return (1);
		}

		free (buf);

		blk += 1;
		ofs += cnt;
	}

	return (0);
}

static
int lec_cat_output_fp (FILE *fp)
{
	unsigned char *crc;
	unsigned long dcrc;
	lec_file_t    lf;
	lec_file_t    *src;

	if (par_lec_cnt == 0) {
		fprintf (stderr, "%s: no files\n", argv0);
		return (1);
	}

	src = &par_lec[0].lf;

	lecf_init (&lf, fp);

	lf.bsize = src->bsize;
	lf.size = src->size;
	lf.wsize = src->wsize;
	lf.bcnt = src->bcnt;
	lf.ecnt = (par_blk_cnt < 65536) ? par_blk_cnt : 65535;
	lf.fsize = src->fsize;
	lf.fcrc = src->fcrc;
	lf.dcrc = src->dcrc;
	memcpy (lf.md5, src->md5, 16);

	lecf_calc_dofs (&lf);
	lecf_calc_eofs (&lf);
	lecf_calc_rofs (&lf);

	crc = lec_read_buf (src->fp, src->dofs, 4 * lf.bcnt);
	if (crc == NULL) {
		fprintf (stderr, "%s: read error (crc)\n", argv0);
		return (1);
	}

	dcrc = lec_get_crc (crc, 4 * lf.bcnt);

	if (lf.dcrc == 0) {
		lf.dcrc = dcrc;
	}
	else {
		if (lf.dcrc != dcrc) {
			fprintf (stderr, "%s: warning: data crc mismatch\n",
				argv0
			);
		}
	}

	if (lecf_write_header (&lf)) {
		fprintf (stderr, "%s: write error (header)\n", argv0);
		return (1);
	}

	if (lecf_write_magic1 (&lf)) {
		fprintf (stderr, "%s: write error (magic1)\n", argv0);
		return (1);
	}

	if (lec_write (lf.fp, lf.dofs, crc, 4 * lf.bcnt) != (4 * lf.bcnt)) {
		fprintf (stderr, "%s: write error (crc)\n", argv0);
		return (1);
	}

	if (lec_cat_output_blocks (&lf)) {
		return (1);
	}

	if (lecf_write_redundant (&lf, crc)) {
		fprintf (stderr, "%s: write error (redundant)\n", argv0);
		return (1);
	}

	free (crc);

	return (0);
}

static
int lec_cat_output (const char *fname)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "wb");
	if (fp == NULL) {
		return (1);
	}

	r = lec_cat_output_fp (fp);

	fclose (fp);

	return (r);
}

static
int lec_cat_input_fp (FILE *fp, const char *fname)
{
	lec_cat_file_t *lc;

	lc = realloc (par_lec, (par_lec_cnt + 1) * sizeof (lec_cat_file_t));
	if (lc == NULL) {
		return (1);
	}

	par_lec = lc;

	lc += par_lec_cnt;

	if (lecf_open (&lc->lf, fp)) {
		return (1);
	}

	par_lec_cnt += 1;

	lc->fname = strdup (fname);

	return (0);
}

static
int lec_cat_input (const char *fname)
{
	FILE *fp;

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		return (1);
	}

	if (lec_cat_input_fp (fp, fname)) {
		fclose (fp);
		return (1);
	}

	return (0);
}

int main (int argc, char **argv)
{
	int        i;
	const char *fname;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg (argv[1], NULL, "--version")) {
			prt_version();
			return (0);
		}
		else if (str_isarg (argv[1], NULL, "--help")) {
			prt_help();
			return (0);
		}
	}

	fname = NULL;

	i = 1;
	while (i < argc) {
		if (str_isarg (argv[i], "-i", "--input")) {
			if ((i + 1) >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n", argv0
				);
				return (1);
			}

			if (lec_cat_input (argv[i + 1])) {
				fprintf (stderr,
					"%s: can't open input file (%s)\n",
					argv0, argv[i + 1]
				);
				return (1);
			}

			i += 1;
		}
		else if (str_isarg (argv[i], "-o", "--output")) {
			if ((i + 1) >= argc) {
				fprintf (stderr,
					"%s: missing output file name\n",
					argv0
				);
				return (1);
			}

			fname = argv[i + 1];

			i += 1;
		}
		else if (str_isarg (argv[i], "-v", "--verbose")) {
			par_verbose = 1;
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (1);
		}
		else {
			if (lec_cat_input (argv[i])) {
				fprintf (stderr,
					"%s: can't open input file (%s)\n",
					argv0, argv[i + 1]
				);
				return (1);
			}
		}

		i += 1;
	}

	if (fname == NULL) {
		if (isatty (fileno (stdout))) {
			fprintf (stderr, "%s: won't write to stdout\n", argv0);
			return (1);
		}
	}

	if (lec_cat_check_files()) {
		return (1);
	}

	if (lec_cat_add_blocks()) {
		return (1);
	}

	if (fname == NULL) {
		if (lec_cat_output_fp (stdout)) {
			return (1);
		}
	}
	else {
		if (lec_cat_output (fname)) {
			return (1);
		}
	}

	return (0);
}
