/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecofile.c                                             *
 * Created:       2005-08-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-08 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecofile.c 48 2006-07-08 11:37:23Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "block.h"
#include "crc.h"
#include "lecofile.h"


int lec_seek (FILE *fp, uint64_t pos)
{
	if (fseeko (fp, pos, SEEK_SET)) {
		return (1);
	}

	return (0);
}

unsigned long lec_read (FILE *fp, uint64_t pos, void *buf, unsigned long cnt)
{
	if (lec_seek (fp, pos)) {
		return (1);
	}

	return (fread (buf, 1, cnt, fp));
}

unsigned long lec_write (FILE *fp, uint64_t pos, const void *buf, unsigned long cnt)
{
	if (lec_seek (fp, pos)) {
		return (1);
	}

	return (fwrite (buf, 1, cnt, fp));
}

unsigned char *lec_read_buf (FILE *fp, uint64_t pos, unsigned long cnt)
{
	unsigned char *buf;

	buf = malloc (cnt);
	if (buf == NULL) {
		return (NULL);
	}

	if (lec_read (fp, pos, buf, cnt) != cnt) {
		free (buf);
		return (NULL);
	}

	return (buf);
}

uint16_t lec_get_uint16 (const void *buf, unsigned i)
{
	uint16_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];

	return (val);
}

uint32_t lec_get_uint32 (const void *buf, unsigned i)
{
	uint32_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];
	val = (val << 8) | tmp[2];
	val = (val << 8) | tmp[3];

	return (val);
}

uint64_t lec_get_uint64 (const void *buf, unsigned i)
{
	uint64_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];
	val = (val << 8) | tmp[2];
	val = (val << 8) | tmp[3];
	val = (val << 8) | tmp[4];
	val = (val << 8) | tmp[5];
	val = (val << 8) | tmp[6];
	val = (val << 8) | tmp[7];

	return (val);
}

void lec_set_uint16 (void *buf, unsigned i, uint16_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 8) & 0xff;
	tmp[1] = val & 0xff;
}

void lec_set_uint32 (void *buf, unsigned i, uint32_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >> 8) & 0xff;
	tmp[3] = val & 0xff;
}

void lec_set_uint64 (void *buf, unsigned i, uint64_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 56) & 0xff;
	tmp[1] = (val >> 48) & 0xff;
	tmp[2] = (val >> 40) & 0xff;
	tmp[3] = (val >> 32) & 0xff;
	tmp[4] = (val >> 24) & 0xff;
	tmp[5] = (val >> 16) & 0xff;
	tmp[6] = (val >> 8) & 0xff;
	tmp[7] = val & 0xff;
}

uint32_t lec_get_crc (const void *buf, unsigned long cnt)
{
	static int     ok = 0;
	static crc32_t crc;

	if (ok == 0) {
		crc32_init (&crc, CRC_CRC32);
		ok = 1;
	}

	crc32_reset (&crc);
	crc32_calc (&crc, buf, cnt);
	crc32_done (&crc, 1);

	return (crc32_get_reg (&crc));
}

void lecf_calc_dofs (lec_file_t *lf)
{
	lf->dofs = ((LECO_HEADER + 15) & ~15ULL) + 16;
}

void lecf_calc_eofs (lec_file_t *lf)
{
	lf->eofs = ((lf->dofs + 4UL * lf->bcnt + 15) & ~15ULL) + 16;
}

void lecf_calc_rofs (lec_file_t *lf)
{
	lf->rofs = ((lf->eofs + lf->ecnt * (lf->size + 8UL) + 15) & ~15ULL) + 16;
}

int lecf_open (lec_file_t *lf, FILE *fp)
{
	unsigned long cnt;
	unsigned char buf[256];

	lf->fp = fp;

	if (lec_read (fp, 0, buf, 8) != 8) {
		return (1);
	}

	lf->magic = lec_get_uint32 (buf, 0);
	if (lf->magic != LECO_MAGIC) {
		return (1);
	}

	lf->version = lec_get_uint16 (buf, 4);
	if (lf->version != 0) {
		return (1);
	}

	lf->hsize = lec_get_uint16 (buf, 6);
	if (lf->hsize < 64) {
		return (1);
	}

	if (lf->hsize > 256) {
		return (1);
	}

	cnt = lf->hsize - 8;

	if (lec_read (fp, 8, buf + 8, cnt) != cnt) {
		return (1);
	}

	lf->bsize = lec_get_uint32 (buf, 8);
	lf->wsize = lec_get_uint32 (buf, 12);
	lf->dofs = lec_get_uint64 (buf, 16);
	lf->eofs = lec_get_uint64 (buf, 24);
	lf->bcnt = lec_get_uint32 (buf, 32);
	lf->ecnt = lec_get_uint32 (buf, 36);
	lf->fsize = lec_get_uint64 (buf, 40);

	lf->size = ((lf->wsize + 7) / 8) * lf->bsize;

	lf->rofs = ((lf->eofs + lf->ecnt * (lf->size + 8UL) + 15) & ~15ULL) + 16;

	memcpy (lf->md5, buf + 48, 16);

	if (lf->hsize >= 80) {
		lf->fcrc = lec_get_uint32 (buf, 68);
		lf->dcrc = lec_get_uint32 (buf, 72);
		lf->hcrc = lec_get_uint32 (buf, 76);

		lec_set_uint32 (buf, 76, 0);

		if (lec_get_crc (buf, lf->hsize) != lf->hcrc) {
			return (1);
		}
	}
	else {
		lf->fcrc = 0;
		lf->dcrc = 0;
		lf->hcrc = 0;
	}

	return (0);
}

void lecf_init (lec_file_t *lf, FILE *fp)
{
	lf->fp = fp;

	lf->magic = LECO_MAGIC;
	lf->version = 0;
	lf->hsize = 80;
	lf->bsize = 0;
	lf->size = 0;
	lf->wsize = 16;
	lf->dofs = 80;
	lf->eofs = 0;
	lf->rofs = 0;
	lf->bcnt = 0;
	lf->ecnt = 0;
	lf->fsize = 0;
	lf->fcrc = 0;
	lf->dcrc = 0;
	lf->hcrc = 0;

	memset (lf->md5, 0, 16);
}

int lecf_write_header (lec_file_t *lf)
{
	unsigned long crc;
	unsigned char buf[128];

	lec_set_uint32 (buf, 0, LECO_MAGIC);
	lec_set_uint16 (buf, 4, 0);
	lec_set_uint16 (buf, 6, 80);
	lec_set_uint32 (buf, 8, lf->bsize);
	lec_set_uint32 (buf, 12, lf->wsize);
	lec_set_uint64 (buf, 16, lf->dofs);
	lec_set_uint64 (buf, 24, lf->eofs);
	lec_set_uint32 (buf, 32, lf->bcnt);
	lec_set_uint32 (buf, 36, lf->ecnt);
	lec_set_uint64 (buf, 40, lf->fsize);

	memcpy (buf + 48, lf->md5, 16);

	lec_set_uint32 (buf, 64, 0);
	lec_set_uint32 (buf, 68, lf->fcrc);
	lec_set_uint32 (buf, 72, lf->dcrc);
	lec_set_uint32 (buf, 76, 0);

	crc = lec_get_crc (buf, 80);
	lec_set_uint32 (buf, 76, crc);

	if (lec_write (lf->fp, 0, buf, 80) != 80) {
		return (1);
	}

	return (0);
}

int lecf_write_magic1 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC1, 8);
	lec_set_uint64 (buf, 8, lf->dofs - 16);

	if (lec_write (lf->fp, lf->dofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_magic2 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC2, 8);
	lec_set_uint64 (buf, 8, lf->eofs - 16);

	if (lec_write (lf->fp, lf->eofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_magic3 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC3, 8);
	lec_set_uint64 (buf, 8, lf->rofs - 16);

	if (lec_write (lf->fp, lf->rofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_redundant (lec_file_t *lf, const void *crc)
{
	unsigned      i;
	unsigned char buf[32];

	if (lecf_write_magic3 (lf)) {
		return (1);
	}

	lec_set_uint32 (buf, 0, lf->bsize);
	lec_set_uint32 (buf, 4, lf->wsize);
	lec_set_uint32 (buf, 8, lf->bcnt);
	lec_set_uint32 (buf, 12, lf->ecnt);
	lec_set_uint64 (buf, 16, lf->eofs);
	lec_set_uint64 (buf, 24, lf->fsize);

	if (lec_write (lf->fp, lf->rofs, buf, 32) != 32) {
		return (1);
	}

	if (crc == NULL) {
		uint64_t dofs, rofs;

		dofs = lf->dofs;
		rofs = lf->rofs + 32;

		for (i = 0; i < lf->bcnt; i++) {
			if (lec_read (lf->fp, dofs, buf, 4) != 4) {
				return (1);
			}

			if (lec_write (lf->fp, rofs, buf, 4) != 4) {
				return (1);
			}

			dofs += 4;
			rofs += 4;
		}
	}
	else {
		unsigned long cnt;

		cnt = 4UL * lf->bcnt;
		if (lec_write (lf->fp, lf->rofs + 32, crc, cnt) != cnt) {
			return (1);
		}
	}

	return (0);
}
