/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecoinfo.c                                             *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecoinfo.c 52 2006-07-10 18:02:11Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "lecofile.h"
#include "util.h"


const char *argv0 = NULL;

static int par_info_all = 1;
static int par_info_dblk = 0;
static int par_info_eblk = 0;
static int par_info_first = 0;
static int par_info_hsize = 0;
static int par_info_md5 = 0;
static int par_info_ecccnt = 0;


typedef struct {
	lec_file_t         lf;

	unsigned long      rbsize;
	unsigned long      rwsize;
	unsigned long      rbcnt;
	unsigned long      recnt;
	unsigned long long reofs;
	unsigned long long rfsize;
} lec_info_t;


static
void prt_version (void)
{
	fputs (
		"lecoinfo version " LECO_VERSION_STR
		" (" LECO_CFG_DATE " " LECO_CFG_TIME ")\n\n"
		"Copyright (C) 2005-2006 by Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_help (void)
{
	fputs (
		"Usage: lecoinfo [options] file\n"
		"  -a, --all              Print all information [default]\n"
		"  -d, --data-blocks      List data blocks\n"
		"  -e, --ecc-blocks       List ECC blocks\n"
		"  -f, --first-block      Print first ECC block index\n"
		"  -h, --header-size      Print header size\n"
		"  -m, --md5              Print MD5 of data file\n"
		"  -n, --ecc-block-count  Print number of ECC blocks\n",
		stdout
	);

	fflush (stdout);
}

static
int lec_info_open (lec_info_t *li, FILE *fp)
{
	unsigned char buf[64];
	lec_file_t    *lf;

	lf = &li->lf;

	if (lecf_open (lf, fp)) {
		return (1);
	}

	lecf_calc_rofs (lf);

	if (lec_read (fp, lf->rofs - 16, buf, 48) != 48) {
		lf->rofs = 0;
	}
	else if (memcmp (buf, LECO_MAGIC3, 8) != 0) {
		lf->rofs = 0;
	}
	else {
		li->rbsize = lec_get_uint32 (buf, 16);
		li->rwsize = lec_get_uint32 (buf, 20);
		li->rbcnt = lec_get_uint32 (buf, 24);
		li->recnt = lec_get_uint32 (buf, 28);
		li->reofs = lec_get_uint64 (buf, 32);
		li->rfsize = lec_get_uint64 (buf, 40);
	}

	return (0);
}

static
int lec_info_all (lec_info_t *li)
{
	unsigned   i;
	lec_file_t *lf;

	lf = &li->lf;

	printf ("magic:            %08lX\n", lf->magic);
	printf ("format version:   %u\n", lf->version);
	printf ("header size:      %u\n", lf->hsize);
	printf ("file size:        %llu\n", lf->fsize);
	printf ("file crc:         %08lx (%08lx)\n", lf->fcrc, lf->dcrc);
	printf ("md5:              ");

	for (i = 0; i < 16; i++) {
		printf ("%02x", lf->md5[i]);
	}

	fputs ("\n", stdout);

	printf ("block size:       %lu words\n", lf->bsize);
	printf ("block size:       %lu bytes\n", lf->size);
	printf ("word size:        %u bits\n", lf->wsize);
	printf ("data crc offset:  0x%08llx\n", lf->dofs);
	printf ("ecc block offset: 0x%08llx\n", lf->eofs);
	printf ("data block count: %u\n", lf->bcnt);
	printf ("ecc block count:  %u\n", lf->ecnt);

	if (lf->rofs != 0) {
		printf ("redundant crcs:   0x%08llx\n", lf->rofs);
		printf ("  block size:     %lu words\n", li->rbsize);
		printf ("  word size:      %lu bits\n", li->rwsize);
		printf ("  data blocks:    %lu\n", li->rbcnt);
		printf ("  ecc blocks:     %lu\n", li->recnt);
		printf ("  ecc offset:     0x%08llx\n", li->reofs);
		printf ("  file size:      %llu\n", li->rfsize);
	}
	else {
		printf ("redundant crcs:   not available\n");
	}

	return (0);
}

static
int lec_info_hsize (lec_info_t *li)
{
	printf ("%u\n", li->lf.hsize);

	return (0);
}

static
int lec_info_ecccnt (lec_info_t *li)
{
	printf ("%u\n", li->lf.ecnt);

	return (0);
}

static
int lec_info_md5 (lec_info_t *li)
{
	unsigned i;

	for (i = 0; i < 16; i++) {
		printf ("%02x", li->lf.md5[i]);
	}

	fputs ("\n", stdout);

	return (0);
}

static
int lec_info_eblk (lec_info_t *li)
{
	unsigned long i;
	unsigned char buf[8];
	unsigned long blk, crc;
	uint64_t      ofs;
	lec_file_t    *lf;

	lf = &li->lf;

	ofs = lf->eofs;

	for (i = 0; i < lf->ecnt; i++) {
		if (lec_read (lf->fp, ofs, buf, 4) != 4) {
			return (1);
		}
		if (lec_read (lf->fp, ofs + lf->size + 4, buf + 4, 4) != 4) {
			return (1);
		}

		blk = lec_get_uint32 (buf, 0);
		crc = lec_get_uint32 (buf, 4);

		printf ("%lu\t%lu\t%08lx\n", i, blk, crc);

		ofs += lf->size + 8;
	}

	return (0);
}

static
int lec_info_dblk (lec_info_t *li)
{
	unsigned long i;
	unsigned char buf[4];
	unsigned long crc;
	uint64_t      ofs;
	lec_file_t    *lf;

	lf = &li->lf;

	ofs = lf->dofs;

	for (i = 0; i < lf->bcnt; i++) {
		if (lec_read (lf->fp, ofs, buf, 4) != 4) {
			return (1);
		}

		crc = lec_get_uint32 (buf, 0);

		printf ("%lu\t%08lx\n", i, crc);

		ofs += 4;
	}

	return (0);
}

static
int lec_info_first (lec_info_t *li)
{
	unsigned char buf[4];
	unsigned long blk;
	lec_file_t    *lf;

	lf = &li->lf;

	if (lec_read (lf->fp, lf->eofs, buf, 4) != 4) {
		return (1);
	}

	blk = lec_get_uint32 (buf, 0);

	printf ("%lu\n", blk);

	return (0);
}

static
int lec_info_fp (FILE *fp)
{
	lec_info_t li;

	if (lec_info_open (&li, fp)) {
		return (1);
	}

	if (par_info_all) {
		lec_info_all (&li);
	}
	else if (par_info_dblk) {
		lec_info_dblk (&li);
	}
	else if (par_info_eblk) {
		lec_info_eblk (&li);
	}
	else if (par_info_first) {
		lec_info_first (&li);
	}
	else if (par_info_hsize) {
		lec_info_hsize (&li);
	}
	else if (par_info_ecccnt) {
		lec_info_ecccnt (&li);
	}
	else if (par_info_md5) {
		lec_info_md5 (&li);
	}

	return (0);
}

static
int lec_info (const char *lec)
{
	int  r;
	FILE *fp;

	fp = fopen (lec, "rb");
	if (fp == NULL) {
		return (1);
	}

	r = lec_info_fp (fp);

	fclose (fp);

	return (r);
}

int main (int argc, char **argv)
{
	int        i, r;
	const char *fname;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg (argv[1], NULL, "--version")) {
			prt_version();
			return (0);
		}
		else if (str_isarg (argv[1], NULL, "--help")) {
			prt_help();
			return (0);
		}
	}

	fname = NULL;

	i = 1;
	while (i < argc) {
		if (str_isarg (argv[i], "-a", "--all")) {
			par_info_all = 1;
		}
		else if (str_isarg (argv[i], "-d", "--data-blocks")) {
			par_info_all = 0;
			par_info_dblk = 1;
		}
		else if (str_isarg (argv[i], "-e", "--ecc-blocks")) {
			par_info_all = 0;
			par_info_eblk = 1;
		}
		else if (str_isarg (argv[i], "-f", "--first-block")) {
			par_info_all = 0;
			par_info_first = 1;
		}
		else if (str_isarg (argv[i], "-h", "--header-size")) {
			par_info_all = 0;
			par_info_hsize = 1;
		}
		else if (str_isarg (argv[i], "-m", "--md5")) {
			par_info_all = 0;
			par_info_md5 = 1;
		}
		else if (str_isarg (argv[i], "-n", "--ecc-block-count")) {
			par_info_all = 0;
			par_info_ecccnt = 1;
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n", argv[0], argv[i]);
			return (1);
		}
		else {
			if (fname == NULL) {
				fname = argv[i];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv[0], argv[i]);
				return (1);
			}
		}

		i += 1;
	}

	if (fname != NULL) {
		r = lec_info (fname);
	}
	else {
		r = 1;
	}

	return (r);
}
